defmodule Mix.Tasks.Pandoc do
  @moduledoc """
  Invokes pandoc with the given args.

  Usage:

      $ mix pandoc TASK_OPTIONS PROFILE PANDOC_ARGS

  Example:

      $ mix pandoc default documents/hello.md -o priv/static/posts/hello.html

  If pandoc is not installed, it is automatically downloaded. Note the
  arguments given to this task will be appended to any configured arguments.

  ## Options

    * `--runtime-config` - load the runtime configuration
      before executing command

  Note flags to control this Mix task must be given before the profile:

      $ mix pandoc --runtime-config default documents/hello.md

  """

  @shortdoc "Invokes pandoc with the profile and args"
  @compile {:no_warn_undefined, Mix}

  use Mix.Task

  @impl true
  def run(args) do
    switches = [runtime_config: :boolean]
    {opts, remaining_args} = OptionParser.parse_head!(args, switches: switches)

    if function_exported?(Mix, :ensure_application!, 1) do
      Mix.ensure_application!(:inets)
      Mix.ensure_application!(:ssl)
    end

    if opts[:runtime_config] do
      Mix.Task.run("app.config")
    else
      Mix.Task.run("loadpaths")
      Application.ensure_all_started(:pandoc)
    end

    Mix.Task.reenable("pandoc")
    install_and_run(remaining_args)
  end

  defp install_and_run([profile | args] = all) do
    case Pandoc.install_and_run(String.to_atom(profile), args) do
      0 -> :ok
      status -> Mix.raise("`mix pandoc #{Enum.join(all, " ")}` exited with #{status}")
    end
  end

  defp install_and_run([]) do
    Mix.raise("`mix pandoc` expects the profile as argument")
  end
end
